!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_add 
!  NAME
!    cadna_add
!
!
!  DESCRIPTION 
!    Define all the functions involving at least one
!    stochastic argument which overload the "+" 
!    operator in a statement such as "a+b" or "+a".
!
!  INPUTS
!    All the functions have one or two arguments of stochastic, 
!    real or integer type in single or double precision. At least 
!    one argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one numerical output of stochastic type. 
!    It can be an array of rank 1 if one input argument is an array 
!    of rank 1. The output is in double precision if, at least,
!    one input is in double precision.
!
!  NOTES
!    If one argumment is an array of rank 1, the arguments cannot
!    be of integer type and must be in double precision.
!
!******
!
module cadna_add
  !
  use cadna_type
  use cadna_cestac
  !
  interface operator (+)
     !       ======================
     module procedure add_st, add_dst, add_st_st, &
          &		add_single_st, add_st_single, add_int_st, &
          &		add_st_int, add_dst_dst, add_dst_double, &
          &		add_double_dst, add_dst_st, add_st_dst, &
          &		add_dst_single, add_single_dst, add_st_double, &
          &		add_double_st, add_int_dst, add_dst_int, &
          !-----------------------------------
          &          add_vdst, add_vdst_vdst, add_dst_vdst, add_vdst_dst, &
          &          add_double_vdst, add_vdst_double, add_vdouble_vdst, &
          &          add_vdst_vdouble, add_vdouble_dst, add_dst_vdouble, &
          !-----------------------------------
          &          add_vvdst_vvdst, add_single_vdst, &
          &          add_int_vdst, add_vdst_int, add_single_vvdst, &
          &          add_vvvdst_vvvdst, &
          &          add_vvvdst_dst, add_dst_vvvdst
  end interface
  !
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  type (single_st) function add_st(a)
    implicit none
    type (single_st), intent(in) :: a
    add_st = a
  end function add_st
  !
  type (double_st) function add_dst(a)
    implicit none
    type (double_st), intent(in) :: a
    add_dst = a
  end function add_dst
  !
  type (single_st) function add_st_st(a,b)
    implicit none
    type (single_st), intent(in) :: a, b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y + b%y
    call rnd_switch
    aux%z = a%z + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_st_st = aux
  end function add_st_st
  !
  type (single_st) function add_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a + b%x
    if (myrandom()) call rnd_switch
    aux%y = a + b%y
    call rnd_switch
    aux%z = a + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_single_st = aux
  end function add_single_st
  !
  type (single_st) function add_st_single(a,b)
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b
    if (myrandom()) call rnd_switch
    aux%y = a%y + b
    call rnd_switch
    aux%z = a%z + b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_st_single = aux
  end function add_st_single
  !
  type (single_st) function add_int_st(a,b)
    implicit none
    integer, intent(in) :: a
    type (single_st), intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a + b%x
    if (myrandom()) call rnd_switch
    aux%y = a + b%y
    call rnd_switch
    aux%z = a + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_int_st = aux
  end function add_int_st
  !
  type (single_st) function add_st_int(a,b)
    implicit none
    type (single_st), intent(in) :: a
    integer, intent(in) :: b
    type (single_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b
    if (myrandom()) call rnd_switch
    aux%y = a%y + b
    call rnd_switch
    aux%z = a%z + b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       aux%acc = nb_significant_digit_st(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_st_int = aux
  end function add_st_int
  !
  type (double_st) function add_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a, b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y + b%y
    call rnd_switch
    aux%z = a%z + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_dst_dst = aux
  end function add_dst_dst
  !
  type (double_st) function add_dst_double(a,b)
    implicit none
    type (double_st), intent(in) ::a
    double precision, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b
    if (myrandom()) call rnd_switch
    aux%y = a%y + b
    call rnd_switch
    aux%z = a%z + b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_dst_double = aux
  end function add_dst_double
  !
  type (double_st) function add_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a + b%x
    if (myrandom()) call rnd_switch
    aux%y = a + b%y
    call rnd_switch
    aux%z = a + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_double_dst = aux
  end function add_double_dst
  !
  type (double_st) function add_dst_st(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y + b%y
    call rnd_switch
    aux%z = a%z + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_dst_st = aux
  end function add_dst_st
  !
  type (double_st) function add_st_dst(a,b)
    implicit none
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b%x
    if (myrandom()) call rnd_switch
    aux%y = a%y + b%y
    call rnd_switch
    aux%z = a%z + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( min(a%acc, b%acc) .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_st_dst = aux
  end function add_st_dst
  !
  type (double_st) function add_single_dst(a,b)
    implicit none
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a + b%x
    if (myrandom()) call rnd_switch
    aux%y = a + b%y
    call rnd_switch
    aux%z = a + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_single_dst = aux
  end function add_single_dst
  !
  type (double_st) function add_dst_single(a,b)
    implicit none
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b
    if (myrandom()) call rnd_switch
    aux%y = a%y + b
    call rnd_switch
    aux%z = a%z + b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_dst_single = aux
  end function add_dst_single
  !
  type (double_st) function add_double_st(a,b)
    implicit none
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a + b%x
    if (myrandom()) call rnd_switch
    aux%y = a + b%y
    call rnd_switch
    aux%z = a + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_st(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_double_st = aux
  end function add_double_st
  !
  type (double_st) function add_st_double(a,b)
    implicit none
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b
    if (myrandom()) call rnd_switch
    aux%y = a%y + b
    call rnd_switch
    aux%z = a%z + b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_st(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_st_double = aux
  end function add_st_double
  !
  type (double_st) function add_int_dst(a,b)
    implicit none
    integer, intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a + b%x
    if (myrandom()) call rnd_switch
    aux%y = a + b%y
    call rnd_switch
    aux%z = a + b%z
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (b%acc.eq.-1) then
          call accuracy_dst(b)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( b%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_int_dst = aux
  end function add_int_dst
  !
  type (double_st) function add_dst_int(a,b)
    implicit none
    type (double_st), intent(in) :: a
    integer, intent(in) :: b
    type (double_st) :: aux
    if (myrandom()) call rnd_switch
    aux%x = a%x + b
    if (myrandom()) call rnd_switch
    aux%y = a%y + b
    call rnd_switch
    aux%z = a%z + b
    aux%acc = -1
    if (cadna_cancellation_tag) then
       if (a%acc.eq.-1) then
          call accuracy_dst(a)
       endif
       aux%acc = nb_significant_digit_dst(aux)
       if ( a%acc .ge. aux%acc + cadna_cancellation_level ) then
          call instability(cadna_cancellation)
       end if
    end if
    add_dst_int = aux
  end function add_dst_int
  !
  !------------------------------------------------------------
  !
  function add_vdst(a)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: add_vdst
    add_vdst%x = a%x
    add_vdst%y = a%y
    add_vdst%z = a%z
    add_vdst%acc = a%acc
  end function add_vdst
  !
  function add_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a,b
    type (double_st), dimension(size(a)) :: add_vdst_vdst
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         add_vdst_vdst(i) = add_dst_dst(a(i), b(i))
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           add_vdst_vdst(i)%x = a(i)%x + b(i)%x 
 !        else
 !           add_vdst_vdst(i)%x = -((-a(i)%x) + (-b(i)%x))
 !        endif
 !        if (myrandom()) then
 !           add_vdst_vdst(i)%y = a(i)%y + b(i)%y
 !           add_vdst_vdst(i)%z = -((-a(i)%z) + (-b(i)%z))
 !        else
 !           add_vdst_vdst(i)%y = -((-a(i)%y) + (-b(i)%y))
 !           add_vdst_vdst(i)%z = a(i)%z + b(i)%z
 !        endif
 !        add_vdst_vdst(i)%acc = -1
 !     enddo
 !   endif
  end function add_vdst_vdst
  !
  function add_dst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: add_dst_vdst
    integer :: k, i
    k = size(b) 
  !  if (cadna_cancellation_tag) then
      do i=1,k
         add_dst_vdst(i) = add_dst_dst(a, b(i))
      enddo
  !  else
  !    do i=1,k
  !       if (myrandom()) then
  !          add_dst_vdst(i)%x = a%x + b(i)%x 
  !       else
  !          add_dst_vdst(i)%x = -((-a%x) + (-b(i)%x))
  !       endif
  !       if (myrandom()) then
  !          add_dst_vdst(i)%y = a%y + b(i)%y
  !          add_dst_vdst(i)%z = -((-a%z) + (-b(i)%z))
  !       else
  !          add_dst_vdst(i)%y = -((-a%y) + (-b(i)%y))
  !          add_dst_vdst(i)%z = a%z + b(i)%z
  !       endif
  !       add_dst_vdst(i)%acc = -1
  !    enddo
  !  endif
  end function add_dst_vdst
  !
  function add_vdst_dst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: add_vdst_dst
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         add_vdst_dst(i) = add_dst_dst(a(i), b)
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           add_vdst_dst(i)%x = a(i)%x + b%x 
 !        else
 !           add_vdst_dst(i)%x = -((-a(i)%x) + (-b%x))
 !        endif
 !        if (myrandom()) then
 !           add_vdst_dst(i)%y = a(i)%y + b%y
 !           add_vdst_dst(i)%z = -((-a(i)%z) + (-b%z))
 !        else
 !           add_vdst_dst(i)%y = -((-a(i)%y) + (-b%y))
 !           add_vdst_dst(i)%z = a(i)%z + b%z
 !        endif
 !        add_vdst_dst(i)%acc = -1
 !     enddo
 !   endif
  end function add_vdst_dst
  !
  function add_double_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b)) :: add_double_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         add_double_vdst(i) = add_double_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_double_vdst(i)%x = a + b(i)%x 
!         else
!            add_double_vdst(i)%x = -((-a) + (-b(i)%x))
!         endif
!         if (myrandom()) then
!            add_double_vdst(i)%y = a + b(i)%y
!            add_double_vdst(i)%z = -((-a) + (-b(i)%z))
!         else
!            add_double_vdst(i)%y = -((-a) + (-b(i)%y))
!            add_double_vdst(i)%z = a + b(i)%z
!         endif
!         add_double_vdst(i)%acc = -1
!      enddo
!    endif
  end function add_double_vdst
  !
  function add_vdst_double(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a)) :: add_vdst_double
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         add_vdst_double(i) = add_dst_double(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_vdst_double(i)%x = a(i)%x + b 
!         else
!            add_vdst_double(i)%x = -((-a(i)%x) + (-b))
!         endif
!         if (myrandom()) then
!            add_vdst_double(i)%y = a(i)%y + b
!            add_vdst_double(i)%z = -((-a(i)%z) + (-b))
!         else
!            add_vdst_double(i)%y = -((-a(i)%y) + (-b))
!            add_vdst_double(i)%z = a(i)%z + b
!         endif
!         add_vdst_double(i)%acc = -1
!      enddo
!    endif
  end function add_vdst_double
  !
  function add_vdouble_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, dimension(:), intent(in) :: a
    type (double_st), dimension(size(b)) :: add_vdouble_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         add_vdouble_vdst(i) = add_double_dst(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_vdouble_vdst(i)%x = a(i) + b(i)%x 
!         else
!            add_vdouble_vdst(i)%x = -((-a(i)) + (-b(i)%x))
!         endif
!         if (myrandom()) then
!            add_vdouble_vdst(i)%y = a(i) + b(i)%y
!            add_vdouble_vdst(i)%z = -((-a(i)) + (-b(i)%z))
!         else
!            add_vdouble_vdst(i)%y = -((-a(i)) + (-b(i)%y))
!            add_vdouble_vdst(i)%z = a(i) + b(i)%z
!         endif
!         add_vdouble_vdst(i)%acc = -1
!      enddo
!    endif
  end function add_vdouble_vdst
  !
  function add_vdst_vdouble(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(a)) :: add_vdst_vdouble
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         add_vdst_vdouble(i) = add_dst_double(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_vdst_vdouble(i)%x = a(i)%x + b(i)
!         else
!            add_vdst_vdouble(i)%x = -((-a(i)%x) + (-b(i)))
!         endif
!         if (myrandom()) then
!            add_vdst_vdouble(i)%y = a(i)%y + b(i)
!            add_vdst_vdouble(i)%z = -((-a(i)%z) + (-b(i)))
!         else
!            add_vdst_vdouble(i)%y = -((-a(i)%y) + (-b(i)))
!            add_vdst_vdouble(i)%z = a(i)%z + b(i)
!         endif
!         add_vdst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function add_vdst_vdouble
  !
  function add_dst_vdouble(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(b)) :: add_dst_vdouble
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         add_dst_vdouble(i) = add_dst_double(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_dst_vdouble(i)%x = a%x + b(i)
!         else
!            add_dst_vdouble(i)%x = -((-a%x) + (-b(i)))
!         endif
!         if (myrandom()) then
!            add_dst_vdouble(i)%y = a%y + b(i)
!            add_dst_vdouble(i)%z = -((-a%z) + (-b(i)))
!         else
!            add_dst_vdouble(i)%y = -((-a%y) + (-b(i)))
!            add_dst_vdouble(i)%z = a%z + b(i)
!         endif
!         add_dst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function add_dst_vdouble
  !
  function add_vdouble_dst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: add_vdouble_dst
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         add_vdouble_dst(i) = add_double_dst(a(i), b)
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           add_vdouble_dst(i)%x = a(i) + b%x 
 !        else
 !           add_vdouble_dst(i)%x = -((-a(i)) + (-b%x))
 !        endif
 !        if (myrandom()) then
 !           add_vdouble_dst(i)%y = a(i) + b%y
 !           add_vdouble_dst(i)%z = -((-a(i)) + (-b%z))
 !        else
 !           add_vdouble_dst(i)%y = -((-a(i)) + (-b%y))
 !           add_vdouble_dst(i)%z = a(i) + b%z
 !        endif
 !        add_vdouble_dst(i)%acc = -1
 !     enddo
 !   endif
  end function add_vdouble_dst
  !
  function add_vvdst_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: a,b
    type (double_st), dimension(size(a,1),size(a,2)) :: add_vvdst_vvdst
    integer :: i, j, n, m
    n=size(a,1)
    m=size(a,2)
    do j=1,m
       do i=1,n
          add_vvdst_vvdst(i,j)=add_dst_dst(a(i,j), b(i,j))
       enddo
    enddo
  end function add_vvdst_vvdst
  !
  function add_single_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    real, intent(in) :: a
    type (double_st), dimension(size(b)) :: add_single_vdst
    integer :: k, i
    k = size(b)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         add_single_vdst(i) = add_single_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_single_vdst(i)%x = a + b(i)%x 
!         else
!            add_single_vdst(i)%x = -((-a) + (-b(i)%x))
!         endif
!         if (myrandom()) then
!            add_single_vdst(i)%y = a + b(i)%y
!            add_single_vdst(i)%z = -((-a) + (-b(i)%z))
!         else
!            add_single_vdst(i)%y = -((-a) + (-b(i)%y))
!            add_single_vdst(i)%z = a + b(i)%z
!         endif
!         add_single_vdst(i)%acc = -1
!      enddo
!    endif
  end function add_single_vdst
  !
  function add_int_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    integer, intent(in) :: a
    type (double_st), dimension(size(b)) :: add_int_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         add_int_vdst(i) = add_int_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            add_int_vdst(i)%x = a + b(i)%x 
!         else
!            add_int_vdst(i)%x = -((-a) + (-b(i)%x))
!         endif
!         if (myrandom()) then
!            add_int_vdst(i)%y = a + b(i)%y
!            add_int_vdst(i)%z = -((-a) + (-b(i)%z))
!         else
!            add_int_vdst(i)%y = -((-a) + (-b(i)%y))
!            add_int_vdst(i)%z = a + b(i)%z
!         endif
!         add_int_vdst(i)%acc = -1
!      enddo
!    endif
  end function add_int_vdst
  !
  function add_vdst_int(a,b)
    implicit none
    integer, intent(in) :: b
    type(double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: add_vdst_int
    add_vdst_int=add_int_vdst(b,a)
  end function add_vdst_int
  !
  function add_single_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    real, intent(in) :: a
    type (double_st), dimension(size(b,1),size(b,2)) :: add_single_vvdst
    integer :: i,j
    do j=1,size(b,2)
       do i=1,size(b,1)
          add_single_vvdst(i,j)=add_single_dst(a,b(i,j))
       enddo
    enddo
  end function add_single_vvdst
  !
  function add_vvvdst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a,b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: add_vvvdst_vvvdst
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             add_vvvdst_vvvdst(i,j,k)=add_dst_dst(a(i,j,k), b(i,j,k))
          enddo
       enddo
    enddo
  end function add_vvvdst_vvvdst
  !
  function add_vvvdst_dst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: add_vvvdst_dst
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             add_vvvdst_dst(i,j,k)=add_dst_dst(a(i,j,k), b)
          enddo
       enddo
    enddo
  end function add_vvvdst_dst
  !
  function add_dst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b,1),size(b,2),size(b,3)) :: add_dst_vvvdst
    add_dst_vvvdst=add_vvvdst_dst(b,a)
  end function add_dst_vvvdst
end module cadna_add
