!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_cestac
!  NAME
!    cadna_cestac
!
!
!  DESCRIPTION
!    Contains the functions and subroutines which compute the accuracy
!    of stochastic variables
!
!******
!
module cadna_cestac  
!
  use cadna_type
!
!****f cadna_cestac/nb_significant_digit 
!  NAME
!    nb_significant_digit
!
!  FUNCTION
!    Functions of the interface "nb_significant_digit" compute
!    the accuracy, i.e. the number of exact significant digits 
!    of any stochastic variable.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    An integer value which is the number of exact significant digits 
!    of the argument.
!
!******
  interface nb_significant_digit
     !       ======================
     module procedure nb_significant_digit_st, nb_significant_digit_dst
  end interface
  !
!****f cadna_cestac/computed_zero
!  NAME
!    computed_zero
!
!  FUNCTION
!    Functions of the interface "computed_zero" determine if 
!    the argument is a computed zero or not. See the documentation
!    for the definition.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    A logical value which is TRUE if the argument is a computed zero 
!    and FALSE if it is not.
!
!******
  interface computed_zero
     !       ======================
     module procedure computed_zero_st, computed_zero_dst
  end interface
  !
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  integer function nb_significant_digit_st(a)
    type (single_st), intent(in) :: a
    real :: x0,y0
    real :: aux=1.0
    logical :: b

    if (aux+1.0e-20.eq.1.0) then
       b = .true.
    else 
       b = .false.
    endif
    call rnd_arr
    x0 = a%x + a%y + a%z
    if (x0.eq.0.0) then
       if (a%x.eq.a%y.and.a%x.eq.a%z) then
          nb_significant_digit_st = 7
       else
          nb_significant_digit_st = 0
       endif
    else
       x0 = 3.0/x0
       y0 = sqrt( (a%x*x0 - 1.)**2 + (a%y*x0 - 1.)**2 + (a%z*x0 - 1.)**2 )
       y0 = y0*1.756549507
       if (y0.le.1.0e-7) then
          nb_significant_digit_st = 7
       else
          y0 = -log10(y0)
          if (y0.ge.0.0) then
             nb_significant_digit_st = y0 +0.5
          else
             nb_significant_digit_st = 0
          end if
       end if
    end if
    if (b) then
       call rnd_moinf
    else
       call rnd_plinf
    endif
  end function nb_significant_digit_st
  !
  integer function nb_significant_digit_dst(a)
    type (double_st), intent(in) :: a
    double precision :: x0, y0
    real :: aux =1.0
    logical :: b
    if (aux+1.0e-20.eq.1.0) then
       b = .true.
    else 
       b = .false.
    endif
    call rnd_arr
    x0 = a%x + a%y + a%z
    if (x0.eq.0.0) then
       if (a%x.eq.a%y.and.a%x.eq.a%z) then
          nb_significant_digit_dst = 15
       else
          nb_significant_digit_dst = 0
       endif
    else
       x0 = 3.D0/x0
       y0 = sqrt( (a%x*x0-1.D0)**2 + (a%y*x0-1.D0)**2 + (a%z*x0-1.D0)**2 )
       y0 = y0*1.7565495069643402D0
       if (y0.le.1.D-15) then
          nb_significant_digit_dst = 15
       else
          y0 = -log10(y0)
          if (y0.ge.0.D0) then
             nb_significant_digit_dst = y0 + 0.5D0
          else
             nb_significant_digit_dst = 0
          end if
       end if
    end if
    if (b) then
       call rnd_moinf
    else
       call rnd_plinf
    endif
  end function nb_significant_digit_dst
  !
  logical function computed_zero_st(a)
    type (single_st), intent(in) :: a
    double precision :: x0
    real :: aux=1.0
    logical :: b
    if (aux+1.0e-20.eq.1.0) then
       b = .true.
    else 
       b = .false.
    endif
    call rnd_arr
    x0 = a%x + a%y + a%z
    if (x0.eq.0.d0) then
       computed_zero_st = .true.
    else
       x0 = 3.D0/x0
       computed_zero_st = ((a%x*x0-1.d0)**2 + (a%y*x0-1.d0)**2 + &
            &  (a%z*x0-1.d0)**2 )*3.0854661704166664D0 .gt. 1.D-1
    endif
    if (b) then 
       call rnd_moinf
    else
       call rnd_plinf
    endif
  end function computed_zero_st
  !
  logical function computed_zero_dst(a)
    type (double_st), intent(in) :: a
    double precision :: x0
    real :: aux=1.0
    logical :: b
    if (aux+1.0e-20.eq.1.0) then
       b = .true.
    else 
       b = .false.
    endif
    call rnd_arr
    x0 = a%x + a%y + a%z
    if (x0.eq.0.d0) then
       computed_zero_dst = .true.
    else
       x0 = 3.D0/x0
       computed_zero_dst = ( (a%x*x0-1.D0)**2 + (a%y*x0-1.D0)**2 +  & 
            &     (a%z*x0-1.D0)**2 )*3.0854661704166664D0 .gt. 1.D-1
    endif
    if (b) then
       call rnd_moinf
    else
       call rnd_plinf
    endif
  end function computed_zero_dst
  !
end module cadna_cestac
!
!****f accuracy_st
!  NAME
!    accuracy_st
!
!  FUNCTION
!    Exactly the same behaviour as the above function "nb_significant_digit".
!  
!  NOTES
!    It is necessary for a simpler numerical debugging. It is simpler 
!    to read information under "gdb" or equivalent on a function, where
!    a breakpoint was put, which is not defined in a module.
!
!******
subroutine accuracy_st(a)
  use cadna_type
  use cadna_cestac
  type (single_st), intent(inout) :: a
  a%acc = nb_significant_digit_st(a)
end subroutine accuracy_st
!
!****f accuracy_dst
!  NAME
!    accuracy_dst
!
!  FUNCTION
!    Exactly the same behaviour as the above function "nb_significant_digit".
!  
!  NOTES
!    It is necessary for a simpler numerical debugging. It is simpler 
!    to read information under "gdb" or equivalent on a function, where
!    a breakpoint was put, which is not defined in a module.
!
!******
subroutine accuracy_dst(a)
  use cadna_type
  use cadna_cestac
  type (double_st), intent(inout) :: a
  a%acc = nb_significant_digit_dst(a)
end subroutine accuracy_dst

