!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_div
!  NAME
!    cadna_div
!
!
!  DESCRIPTION
!    Define all the functions involving at least one argument
!    of stochastic type which overload the "/" operator
!    in a statement such as "a/b".
!
!  INPUTS
!    All the functions have one or two arguments of stochastic, 
!    real or integer type in single or double precision. At least 
!    one argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one numerical output of stochastic type. 
!    It can be an array of rank 1 if one input argument is an array 
!    of rank 1. The output is in double precision if, at least,
!    one input is in double precision.
!
!  NOTES
!    If one argumment is an array of rank 1, the arguments cannot
!    be of integer type and must be in double precision.

!
!******
!
module cadna_div
  !
  use cadna_type
  use cadna_cestac
  !
  interface operator (/)
     !       ======================
     module procedure div_st_st, &
          &		div_single_st, div_st_single, div_int_st, &
          &		div_st_int, div_dst_dst, div_dst_double, &
          &		div_double_dst, div_dst_st, div_st_dst, &
          &		div_dst_single, div_single_dst, div_st_double, &
          &		div_double_st, div_int_dst, div_dst_int, &
          !--------------------------------------------------------------
          &          div_vdst_vdst, div_dst_vdst, div_vdst_dst, div_double_vdst, &
          &          div_vdst_double, div_vdouble_vdst, div_vdst_vdouble, &
          &          div_vdouble_dst, div_dst_vdouble, &
          !--------------------------------------------------------------
          &          div_dst_vvdst, div_vvdst_dst, div_double_vvdst, div_vvdst_double, &
          !--------------------------------------------------------------
          &          div_vdst_int, div_vdst_single, &
          &          div_vvvdst_vvvdst, &
          &          div_vvvdst_single 
  end interface
  !
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  type (single_st) function div_st_st(a,b)
    implicit none
    type (single_st), intent(in) :: a,b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_st_st%x = a%x / b%x
    if (myrandom()) call rnd_switch
    div_st_st%y = a%y / b%y
    call rnd_switch
    div_st_st%z = a%z / b%z
    div_st_st%acc = -1
  end function div_st_st
  !
  type (single_st) function div_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_single_st%x = a / b%x
    if (myrandom()) call rnd_switch
    div_single_st%y = a / b%y
    call rnd_switch
    div_single_st%z = a / b%z
    div_single_st%acc = -1
  end function div_single_st
  !
  type (single_st) function div_st_single(a,b)
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    if (myrandom()) call rnd_switch
    div_st_single%x = a%x / b
    if (myrandom()) call rnd_switch
    div_st_single%y = a%y / b
    call rnd_switch
    div_st_single%z = a%z / b
    div_st_single%acc = -1
  end function div_st_single
  !
  type (single_st) function div_int_st(a,b)
    implicit none
    type (single_st), intent(in) :: b
    integer, intent(in) :: a
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_int_st%x = a / b%x
    if (myrandom()) call rnd_switch
    div_int_st%y = a / b%y
    call rnd_switch
    div_int_st%z = a / b%z
    div_int_st%acc = -1
  end function div_int_st
  !
  type (single_st) function div_st_int(a,b)
    implicit none
    type (single_st), intent(in) :: a
    integer, intent(in) :: b
    if (myrandom()) call rnd_switch
    div_st_int%x = a%x / b
    if (myrandom()) call rnd_switch
    div_st_int%y = a%y / b
    call rnd_switch
    div_st_int%z = a%z / b
    div_st_int%acc = -1
  end function div_st_int
  !
  type (double_st) function div_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a,b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_dst_dst%x = a%x / b%x
    if (myrandom()) call rnd_switch
    div_dst_dst%y = a%y / b%y
    call rnd_switch
    div_dst_dst%z = a%z / b%z
    div_dst_dst%acc = -1
  end function div_dst_dst
  !
  type (double_st) function div_dst_double(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    if (myrandom()) call rnd_switch
    div_dst_double%x = a%x / b
    if (myrandom()) call rnd_switch
    div_dst_double%y = a%y / b
    call rnd_switch
    div_dst_double%z = a%z / b
    div_dst_double%acc = -1
  end function div_dst_double
  !
  type (double_st) function div_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_double_dst%x = a / b%x
    if (myrandom()) call rnd_switch
    div_double_dst%y = a / b%y
    call rnd_switch
    div_double_dst%z = a / b%z
    div_double_dst%acc = -1
  end function div_double_dst
  !
  type (double_st) function div_dst_st(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_dst_st%x = a%x / b%x
    if (myrandom()) call rnd_switch
    div_dst_st%y = a%y / b%y
    call rnd_switch
    div_dst_st%z = a%z / b%z
    div_dst_st%acc = -1
  end function div_dst_st
  !
  type (double_st) function div_st_dst(a,b)
    implicit none
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_st_dst%x = a%x / b%x
    if (myrandom()) call rnd_switch
    div_st_dst%y = a%y / b%y
    call rnd_switch
    div_st_dst%z = a%z / b%z
    div_st_dst%acc = -1
  end function div_st_dst
  !
  type (double_st) function div_single_dst(a,b)
    implicit none
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_single_dst%x = a / b%x
    if (myrandom()) call rnd_switch
    div_single_dst%y = a / b%y
    call rnd_switch
    div_single_dst%z = a / b%z
    div_single_dst%acc = -1
  end function div_single_dst
  !
  type (double_st) function div_dst_single(a,b)
    implicit none
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    if (myrandom()) call rnd_switch
    div_dst_single%x = a%x / b
    if (myrandom()) call rnd_switch
    div_dst_single%y = a%y / b
    call rnd_switch
    div_dst_single%z = a%z / b
    div_dst_single%acc = -1
  end function div_dst_single
  !
  type (double_st) function div_double_st(a,b)
    implicit none
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_st(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_double_st%x = a / b%x
    if (myrandom()) call rnd_switch
    div_double_st%y = a / b%y
    call rnd_switch
    div_double_st%z = a / b%z
    div_double_st%acc = -1
  end function div_double_st
  !
  type (double_st) function div_st_double(a,b)
    implicit none
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    if (myrandom()) call rnd_switch
    div_st_double%x = a%x / b
    if (myrandom()) call rnd_switch
    div_st_double%y = a%y / b
    call rnd_switch
    div_st_double%z = a%z / b
    div_st_double%acc = -1
  end function div_st_double
  !
  type (double_st) function div_int_dst(a,b)
    implicit none
    integer, intent(in) :: a
    type (double_st), intent(in) :: b
    if (cadna_division_tag) then
       if (b%acc.eq.-1) call accuracy_dst(b) 
       if (b%acc.eq.0) then
          call instability(cadna_division)
       end if
    end if
    if (myrandom()) call rnd_switch
    div_int_dst%x = a / b%x
    if (myrandom()) call rnd_switch
    div_int_dst%y = a / b%y
    call rnd_switch
    div_int_dst%z = a / b%z
    div_int_dst%acc = -1
  end function div_int_dst
  !
  type (double_st) function div_dst_int(a,b)
    implicit none
    type (double_st), intent(in) :: a
    integer, intent(in) :: b
    if (myrandom()) call rnd_switch
    div_dst_int%x = a%x / b
    if (myrandom()) call rnd_switch
    div_dst_int%y = a%y / b
    call rnd_switch
    div_dst_int%z = a%z / b
    div_dst_int%acc = -1
  end function div_dst_int
!
!------------------------------------------------------------
!
  function div_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a,b
    type (double_st), dimension(size(a)) :: div_vdst_vdst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         div_vdst_vdst(i) = div_dst_dst(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_vdst_vdst(i)%x = a(i)%x / b(i)%x 
!         else
!            div_vdst_vdst(i)%x = -((-a(i)%x) / b(i)%x)
!         endif
!         if (myrandom()) then
!            div_vdst_vdst(i)%y = a(i)%y / b(i)%y
!            div_vdst_vdst(i)%z = -((-a(i)%z) / b(i)%z)
!         else
!            div_vdst_vdst(i)%y = -((-a(i)%y) / b(i)%y)
!            div_vdst_vdst(i)%z = a(i)%z / b(i)%z
!         endif
!         div_vdst_vdst(i)%acc = -1
!      enddo
!    endif
  end function div_vdst_vdst
  !
  function div_dst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: div_dst_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         div_dst_vdst(i) = div_dst_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_dst_vdst(i)%x = a%x / b(i)%x 
!         else
!            div_dst_vdst(i)%x = -((-a%x) / b(i)%x)
!         endif
!         if (myrandom()) then
!            div_dst_vdst(i)%y = a%y / b(i)%y
!            div_dst_vdst(i)%z = -((-a%z) / b(i)%z)
!         else
!            div_dst_vdst(i)%y = -((-a%y) / b(i)%y)
!            div_dst_vdst(i)%z = a%z / b(i)%z
!         endif
!         div_dst_vdst(i)%acc = -1
!      enddo
!    endif
  end function div_dst_vdst
  !
  function div_vdst_dst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: div_vdst_dst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         div_vdst_dst(i) = div_dst_dst(a(i), b)
      enddo  
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_vdst_dst(i)%x = a(i)%x / b%x 
!         else
!            div_vdst_dst(i)%x = -((-a(i)%x) / b%x)
!         endif
!         if (myrandom()) then
!            div_vdst_dst(i)%y = a(i)%y / b%y
!            div_vdst_dst(i)%z = -((-a(i)%z) / b%z)
!         else
!            div_vdst_dst(i)%y = -((-a(i)%y) / b%y)
!            div_vdst_dst(i)%z = a(i)%z / b%z
!         endif
!         div_vdst_dst(i)%acc = -1
!      enddo
!    endif
  end function div_vdst_dst
  !
  function div_double_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b)) :: div_double_vdst
    integer :: k, i
    k = size(b)
!   if (cadna_cancellation_tag) then
      do i=1,k
         div_double_vdst(i) = div_double_dst(a, b(i))
      enddo
!   else 
!     do i=1,k
!         if (myrandom()) then
!            div_double_vdst(i)%x = a/ b(i)%x 
!         else
!            div_double_vdst(i)%x = -((-a) / b(i)%x)
!         endif
!         if (myrandom()) then
!            div_double_vdst(i)%y = a / b(i)%y
!            div_double_vdst(i)%z = -((-a) / b(i)%z)
!         else
!            div_double_vdst(i)%y = -((-a) / b(i)%y)
!            div_double_vdst(i)%z = a / b(i)%z
!         endif
!         div_double_vdst(i)%acc = -1
!      enddo
!    endif
  end function div_double_vdst
  !
  function div_vdst_double(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a)) :: div_vdst_double
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         div_vdst_double(i) = div_dst_double(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_vdst_double(i)%x = a(i)%x / b 
!         else
!            div_vdst_double(i)%x = -((-a(i)%x) / b)
!         endif
!         if (myrandom()) then
!            div_vdst_double(i)%y = a(i)%y / b
!            div_vdst_double(i)%z = -((-a(i)%z) / b)
!         else
!            div_vdst_double(i)%y = -((-a(i)%y) / b)
!            div_vdst_double(i)%z = a(i)%z / b
!         endif
!         div_vdst_double(i)%acc = -1
!      enddo
!    endif
  end function div_vdst_double
  !
  function div_vdouble_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, dimension(:), intent(in) :: a
    type (double_st), dimension(size(b)) :: div_vdouble_vdst
    integer :: k, i
    k = size(b)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         div_vdouble_vdst(i) = div_double_dst(a(i), b(i))
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           div_vdouble_vdst(i)%x = a(i) / b(i)%x 
 !        else
 !           div_vdouble_vdst(i)%x = -((-a(i)) / b(i)%x)
 !        endif
 !        if (myrandom()) then
 !           div_vdouble_vdst(i)%y = a(i) / b(i)%y
 !           div_vdouble_vdst(i)%z = -((-a(i)) / b(i)%z)
 !        else
 !           div_vdouble_vdst(i)%y = -((-a(i)) / b(i)%y)
 !           div_vdouble_vdst(i)%z = a(i) / b(i)%z
 !        endif
 !        div_vdouble_vdst(i)%acc = -1
 !     enddo
 !    endif
  end function div_vdouble_vdst
  !
  function div_vdst_vdouble(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(a)) :: div_vdst_vdouble
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         div_vdst_vdouble(i) = div_dst_double(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_vdst_vdouble(i)%x = a(i)%x / b(i)
!         else
!            div_vdst_vdouble(i)%x = -((-a(i)%x) / b(i))
!         endif
!         if (myrandom()) then
!            div_vdst_vdouble(i)%y = a(i)%y / b(i)
!            div_vdst_vdouble(i)%z = -((-a(i)%z) / b(i))
!         else
!            div_vdst_vdouble(i)%y = -((-a(i)%y) / b(i))
!            div_vdst_vdouble(i)%z = a(i)%z / b(i)
!         endif
!         div_vdst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function div_vdst_vdouble
  !
  function div_dst_vdouble(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: div_dst_vdouble
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         div_dst_vdouble(i) = div_dst_double(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_dst_vdouble(i)%x = a%x / b(i) 
!         else
!            div_dst_vdouble(i)%x = -((-a%x) / b(i))
!         endif
!         if (myrandom()) then
!            div_dst_vdouble(i)%y = a%y / b(i)
!            div_dst_vdouble(i)%z = -((-a%z) / b(i))
!         else
!            div_dst_vdouble(i)%y = -((-a%y) / b(i))
!            div_dst_vdouble(i)%z = a%z / b(i)
!         endif
!         div_dst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function div_dst_vdouble
  !
  function div_vdouble_dst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: div_vdouble_dst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         div_vdouble_dst(i) = div_double_dst(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            div_vdouble_dst(i)%x = a(i) / b%x 
!         else
!            div_vdouble_dst(i)%x = -((-a(i)) / b%x)
!         endif
!         if (myrandom()) then
!            div_vdouble_dst(i)%y = a(i) / b%y
!            div_vdouble_dst(i)%z = -((-a(i)) / b%z)
!         else
!            div_vdouble_dst(i)%y = -((-a(i)) / b%y)
!            div_vdouble_dst(i)%z = a(i) / b%z
!         endif
!         div_vdouble_dst(i)%acc = -1
!      enddo
!    endif
  end function div_vdouble_dst
  !
  function div_dst_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2)) :: div_dst_vvdst
    integer i, n
    n=size(b,2)
    do i=1,n
       div_dst_vvdst(:,i)=div_dst_vdst(a,div_dst_vvdst(:,i))
    enddo
  end function div_dst_vvdst
  !
  function div_vvdst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: b
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2)) :: div_vvdst_dst
    div_vvdst_dst=div_dst_vvdst(b,a)
  end function div_vvdst_dst
  !
  function div_double_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2)) :: div_double_vvdst
    integer i, n
    n=size(b,2)
    do i=1,n
       div_double_vvdst(:,i)=div_double_vdst(a,div_double_vvdst(:,i))
    enddo
  end function div_double_vvdst
  !
  function div_vvdst_double(a,b)
    implicit none
    double precision, intent(in) :: b
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2)) :: div_vvdst_double
    div_vvdst_double=div_double_vvdst(b,a)
  end function div_vvdst_double
  !
  function div_vdst_int(a,b)
    implicit none
    integer, intent(in) :: b
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: div_vdst_int
    integer :: i,n
    n=size(a)
    do i=1,n
       div_vdst_int(i)=div_dst_int(a(i),b)
    enddo
  end function div_vdst_int
  !
  function div_vdst_single(a,b)
    implicit none
    real, intent(in) :: b
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: div_vdst_single
    integer :: i,n
    n=size(a)
    do i=1,n
       div_vdst_single(i)=div_dst_single(a(i),b)
    enddo
  end function div_vdst_single
  !
  function div_vvvdst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a, b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: div_vvvdst_vvvdst
    integer :: i,j,k
        do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             div_vvvdst_vvvdst(i,j,k)=div_dst_dst(a(i,j,k),b(i,j,k))
          enddo
       enddo
    enddo
  end function div_vvvdst_vvvdst
  !
  function div_vvvdst_single(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: a
    real, intent(in) :: b
    type (double_st), dimension(size(a,1),size(a,2),size(a,3)) :: div_vvvdst_single
    integer :: i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             div_vvvdst_single(i,j,k)=div_dst_single(a(i,j,k),b)
          enddo
       enddo
    enddo
  end function div_vvvdst_single
  !
end module cadna_div
