!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_math
!  NAME
!    cadna_math
!
!  DESCRIPTION
!    Contains the overloading  of the classical mathematical functions 
!    of FORTRAN 77 for the stochastic types.
!
!******
module cadna_math
  !
  use cadna_type
  use cadna_cestac
  !
!****f cadna_math/"**"
!  NAME
!    operator "**"
!
!  FUNCTION
!    Functions of the interface "**" are the equivalent of classical
!    "**" function for stochastic types.
!
!  INPUTS
!    All the functions have two arguments of stochastic, real or 
!    integer type in single or double precision. At least one 
!    argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one numerical output of stochastic type. 
!    The output is in double precision if, at least, one input is 
!    in double precision.
!
!******
  interface operator (**)
     !       ======================= 
     module procedure pow_st_int, pow_st_st, pow_st_single, &
          &          pow_single_st, pow_dst_int, pow_dst_dst, pow_dst_double, &
          &          pow_double_dst
  end interface
  !
!****f cadna_math/log
!  NAME
!    log
!
!  FUNCTION
!    Functions of the interface "log" are the equivalent of classical
!    LOG function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface log
     !       ======================= 
     module procedure log_st, log_dst
  end interface
  !
!****f cadna_math/sqrt
!  NAME
!    sqrt
!
!  FUNCTION
!    Functions of the interface "sqrt" are the equivalent of classical
!    SQRT function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!    the argument can be an array of rank 1 only in double precision.
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface sqrt
     !       ======================= 
     module procedure sqrt_st, sqrt_dst, sqrt_vdst
  end interface
  !
!****f cadna_math/exp
!  NAME
!    exp
!
!  FUNCTION
!    Functions of the interface "exp" are the equivalent of classical
!    EXP function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface exp
     !       ======================= 
     module procedure exp_st, exp_dst
  end interface
  !
!****f cadna_math/erf
!  NAME
!    erf
!
!  FUNCTION
!    Functions of the interface "erf" are the equivalent of classical
!    ERF function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface erf
     !       ======================= 
     module procedure erf_st, erf_dst
  end interface
  !
  !
!****f cadna_math/derf
!  NAME
!    derf
!
!  FUNCTION
!    Functions of the interface "derf" are the equivalent of classical
!    DERF function for the stochastic double precision type.
!
!  INPUTS
!    a: double_st
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface derf
     !       ======================= 
     module procedure derf_dst
  end interface
  !
!****f cadna_math/sin
!  NAME
!    sin
!
!  FUNCTION
!    Functions of the interface "sin" are the equivalent of classical
!    SIN function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface sin
     !       ======================= 
     module procedure sin_st, sin_dst
  end interface
  !
!****f cadna_math/cos
!  NAME
!    cos
!
!  FUNCTION
!    Functions of the interface "cos" are the equivalent of classical
!    COS function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface cos
     !       ======================= 
     module procedure cos_st, cos_dst
  end interface
  !
!****f cadna_math/log10
!  NAME
!    log10
!
!  FUNCTION
!    Functions of the interface "log10" are the equivalent of classical
!    LOG10 function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface log10
     !       ======================= 
     module procedure log10_st, log10_dst
  end interface
  !
!****f cadna_math/asin
!  NAME
!    asin
!
!  FUNCTION
!    Functions of the interface "asin" are the equivalent of classical
!    ASIN function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface asin
     !       ======================= 
     module procedure asin_st, asin_dst
  end interface
  !
!****f cadna_math/acos
!  NAME
!    acos
!
!  FUNCTION
!    Functions of the interface "acos" are the equivalent of classical
!    ACOS function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface acos
     !       ======================= 
     module procedure acos_st, acos_dst
  end interface
  !
!****f cadna_math/atan
!  NAME
!    atan
!
!  FUNCTION
!    Functions of the interface "atan" are the equivalent of classical
!    ATAN function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface atan
     !       ======================= 
     module procedure atan_st, atan_dst
  end interface
  !
!****f cadna_math/atanh
!  NAME
!    atanh
!
!  FUNCTION
!    Functions of the interface "atanh" are the equivalent of classical
!    ATANH function for stochastic types.
!
!  INPUTS
!    a: single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface atanh
     !       ======================= 
     module procedure atanh_st, atanh_dst
  end interface
  !
!****f cadna_math/atan2
!  NAME
!    atan2
!
!  FUNCTION
!    Functions of the interface "atan2" are the equivalent of classical
!    ATAN2 function for stochastic types.
!
!  INPUTS
!    a, b : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input arguments. 
!
!******
  interface atan2
     !       ======================= 
     module procedure atan2_st, atan2_dst
  end interface
  !
!****f cadna_math/sinh
!  NAME
!    sinh
!
!  FUNCTION
!    Functions of the interface "sinh" are the equivalent of classical
!    SINH function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface sinh
     !       ======================= 
     module procedure sinh_st, sinh_dst
  end interface
  !
!****f cadna_math/cosh
!  NAME
!    cosh
!
!  FUNCTION
!    Functions of the interface "cosh" are the equivalent of classical
!    COSH function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface cosh
     !       ======================= 
     module procedure cosh_st, cosh_dst
  end interface
  !
!****f cadna_math/tanh
!  NAME
!    tanh
!
!  FUNCTION
!    Functions of the interface "tanh" are the equivalent of classical
!    TANH function for stochastic types.
!
!  INPUTS
!    a : single_st (or double_st)
!
!  RETURN VALUE
!    A numerical variable of the same type as the input argument. 
!
!******
  interface tanh
     !       ======================= 
     module procedure tanh_st, tanh_dst
  end interface
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  type (single_st) function pow_st_int(a,n) 
    implicit none
    type (single_st), intent(in) :: a
    integer, intent(in) :: n
    if (cadna_power_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_st_int%x = a%x ** n
    if (myrandom()) call rnd_switch
    pow_st_int%y = a%y ** n
    call rnd_switch
    pow_st_int%z = a%z ** n
    pow_st_int%acc = -1
  end function pow_st_int
  !
  type (single_st) function pow_st_st(a,b) 
    implicit none
    type (single_st), intent(in) :: a, b
    if (cadna_power_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (b%acc .eq. -1) call accuracy_st(b)
       if (a%acc.eq.0.or.b%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_st_st%x = a%x ** b%x
    if (myrandom()) call rnd_switch
    pow_st_st%y = a%y ** b%y
    call rnd_switch
    pow_st_st%z = a%z ** b%z
    pow_st_st%acc = -1
  end function pow_st_st
  !
  type (single_st) function pow_st_single(a,b) 
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    if (cadna_power_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_st_single%x = a%x ** b
    if (myrandom()) call rnd_switch
    pow_st_single%y = a%y ** b
    call rnd_switch
    pow_st_single%z = a%z ** b
    pow_st_single%acc = -1
  end function pow_st_single
  !
  type (single_st) function pow_single_st(a,b) 
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    if (cadna_power_tag) then
       if (b%acc .eq. -1) call accuracy_st(b)
       if (b%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_single_st%x = a ** b%x
    if (myrandom()) call rnd_switch
    pow_single_st%y = a ** b%y
    call rnd_switch
    pow_single_st%z = a ** b%z
    pow_single_st%acc = -1
  end function pow_single_st
  !
  type (double_st) function pow_dst_int(a,n) 
    implicit none
    type (double_st), intent(in) :: a
    integer, intent(in) :: n
    if (cadna_power_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_dst_int%x = a%x ** n
    if (myrandom()) call rnd_switch
    pow_dst_int%y = a%y ** n
    call rnd_switch
    pow_dst_int%z = a%z ** n
    pow_dst_int%acc = -1
  end function pow_dst_int
  !
  type (double_st) function pow_dst_dst(a,b) 
    implicit none
    type (double_st), intent(in) :: a, b
    if (cadna_power_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (b%acc .eq. -1) call accuracy_dst(b)
       if (a%acc.eq.0.or.b%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_dst_dst%x = a%x ** b%x
    if (myrandom()) call rnd_switch
    pow_dst_dst%y = a%y ** b%y
    call rnd_switch
    pow_dst_dst%z = a%z ** b%z
    pow_dst_dst%acc = -1
  end function pow_dst_dst
  !
  type (double_st) function pow_dst_double(a,b) 
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    if (cadna_power_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_dst_double%x = a%x ** b
    if (myrandom()) call rnd_switch
    pow_dst_double%y = a%y ** b
    call rnd_switch
    pow_dst_double%z = a%z ** b
    pow_dst_double%acc = -1
  end function pow_dst_double
  !
  type (double_st) function pow_double_dst(a,b) 
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    if (cadna_power_tag) then
       if (b%acc .eq. -1) call accuracy_dst(b)
       if (b%acc.eq.0) then
          call instability(cadna_power)
       end if
    end if
    if (myrandom()) call rnd_switch
    pow_double_dst%x = a ** b%x
    if (myrandom()) call rnd_switch
    pow_double_dst%y = a ** b%y
    call rnd_switch
    pow_double_dst%z = a ** b%z
    pow_double_dst%acc = -1
  end function pow_double_dst
  !
  type (single_st) function log_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    log_st%x = log(a%x)
    if (myrandom()) call rnd_switch
    log_st%y = log(a%y)
    call rnd_switch
    log_st%z = log(a%z)
    log_st%acc = -1
  end function log_st
  !
  type (double_st) function log_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    log_dst%x = log(a%x)
    if (myrandom()) call rnd_switch
    log_dst%y = log(a%y)
    call rnd_switch
    log_dst%z = log(a%z)
    log_dst%acc = -1
  end function log_dst
  !
  type (single_st) function sqrt_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    sqrt_st%x = sqrt(a%x)
    if (myrandom()) call rnd_switch
    sqrt_st%y = sqrt(a%y)
    call rnd_switch
    sqrt_st%z = sqrt(a%z)
    sqrt_st%acc = -1
  end function sqrt_st
  !
  type (double_st) function sqrt_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    sqrt_dst%x = sqrt(a%x)
    if (myrandom()) call rnd_switch
    sqrt_dst%y = sqrt(a%y)
    call rnd_switch
    sqrt_dst%z = sqrt(a%z)
    sqrt_dst%acc = -1
  end function sqrt_dst
  !
  !-------------------------------------------------------
  !
  function sqrt_vdst(a) 
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: sqrt_vdst
    integer :: i
    do i=1, size(a)
       sqrt_vdst(i) = sqrt_dst(a(i))
    enddo
  end function sqrt_vdst
  !
  type (single_st) function exp_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    exp_st%x = exp(a%x)
    if (myrandom()) call rnd_switch
    exp_st%y = exp(a%y)
    call rnd_switch
    exp_st%z = exp(a%z)
    exp_st%acc = -1
  end function exp_st
  !
  type (double_st) function exp_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0.) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    exp_dst%x = exp(a%x)
    if (myrandom()) call rnd_switch
    exp_dst%y = exp(a%y)
    call rnd_switch
    exp_dst%z = exp(a%z)
    exp_dst%acc = -1
  end function exp_dst
  !
  type (single_st) function erf_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    erf_st%x = erf(a%x)
    if (myrandom()) call rnd_switch
    erf_st%y = erf(a%y)
    call rnd_switch
    erf_st%z = erf(a%z)
    erf_st%acc = -1
  end function erf_st
  !
  type (double_st) function erf_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0.) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    erf_dst%x = erf(a%x)
    if (myrandom()) call rnd_switch
    erf_dst%y = erf(a%y)
     call rnd_switch
    erf_dst%z = erf(a%z)
    erf_dst%acc = -1
  end function erf_dst
!
  type (double_st) function derf_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0.) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    derf_dst%x = derf(a%x)
    if (myrandom()) call rnd_switch
    derf_dst%y = derf(a%y)
    call rnd_switch
    derf_dst%z = derf(a%z)
    derf_dst%acc = -1
  end function derf_dst
  !
  type (single_st) function sin_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    sin_st%x = sin(a%x)
    if (myrandom()) call rnd_switch
    sin_st%y = sin(a%y)
    call rnd_switch
    sin_st%z = sin(a%z)
    sin_st%acc = -1
  end function sin_st
  !
  type (double_st) function sin_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    sin_dst%x = sin(a%x)
    if (myrandom()) call rnd_switch
    sin_dst%y = sin(a%y)
    call rnd_switch
    sin_dst%z = sin(a%z)
    sin_dst%acc = -1
  end function sin_dst
  !
  type (single_st) function cos_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    cos_st%x = cos(a%x)
    if (myrandom()) call rnd_switch
    cos_st%y = cos(a%y)
    call rnd_switch
    cos_st%z = cos(a%z)
    cos_st%acc = -1
  end function cos_st
  !
  type (double_st) function cos_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    cos_dst%x = cos(a%x)
    if (myrandom()) call rnd_switch
    cos_dst%y = cos(a%y)
    call rnd_switch
    cos_dst%z = cos(a%z)
    cos_dst%acc = -1
  end function cos_dst
  !
  type (single_st) function log10_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    log10_st%x = log10(a%x)
    if (myrandom()) call rnd_switch
    log10_st%y = log10(a%y)
    call rnd_switch
    log10_st%z = log10(a%z)
    log10_st%acc = -1
  end function log10_st
  !
  type (double_st) function log10_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    log10_dst%x = log10(a%x)
    if (myrandom()) call rnd_switch
    log10_dst%y = log10(a%y)
    call rnd_switch
    log10_dst%z = log10(a%z)
    log10_dst%acc = -1
  end function log10_dst
  !
  type (single_st) function asin_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    asin_st%x = asin(a%x)
    if (myrandom()) call rnd_switch
    asin_st%y = asin(a%y)
    call rnd_switch
    asin_st%z = asin(a%z)
    asin_st%acc = -1
  end function asin_st
  !
  type (double_st) function asin_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    asin_dst%x = asin(a%x)
    if (myrandom()) call rnd_switch
    asin_dst%y = asin(a%y)
    call rnd_switch
    asin_dst%z = asin(a%z)
    asin_dst%acc = -1
  end function asin_dst
  !
  type (single_st) function acos_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    acos_st%x = acos(a%x)
    if (myrandom()) call rnd_switch
    acos_st%y = acos(a%y)
    call rnd_switch
    acos_st%z = acos(a%z)
    acos_st%acc = -1
  end function acos_st
  !
  type (double_st) function acos_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    acos_dst%x = acos(a%x)
    if (myrandom()) call rnd_switch
    acos_dst%y = acos(a%y)
    call rnd_switch
    acos_dst%z = acos(a%z)
    acos_dst%acc = -1
  end function acos_dst
  !
  type (single_st) function atan_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    atan_st%x = atan(a%x)
    if (myrandom()) call rnd_switch
    atan_st%y = atan(a%y)
    call rnd_switch
    atan_st%z = atan(a%z)
    atan_st%acc = -1
  end function atan_st
  !
  type (double_st) function atan_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    atan_dst%x = atan(a%x)
    if (myrandom()) call rnd_switch
    atan_dst%y = atan(a%y)
    call rnd_switch
    atan_dst%z = atan(a%z)
    atan_dst%acc = -1
  end function atan_dst
  !
  type (single_st) function atanh_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    atanh_st%x = atanh(a%x)
    if (myrandom()) call rnd_switch
    atanh_st%y = atanh(a%y)
     call rnd_switch
    atanh_st%z = atanh(a%z)
    atanh_st%acc = -1
  end function atanh_st
  !
  type (double_st) function atanh_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    atanh_dst%x = atanh(a%x)
    if (myrandom()) call rnd_switch
    atanh_dst%y = atanh(a%y)
    call rnd_switch
    atanh_dst%z = atanh(a%z)
    atanh_dst%acc = -1
  end function atanh_dst
  !
  type (single_st) function atan2_st(a,b) 
    implicit none
    type (single_st), intent(in) :: a, b
    if (myrandom()) call rnd_switch
    atan2_st%x = atan2(a%x,b%x)
    if (myrandom()) call rnd_switch
    atan2_st%y = atan2(a%y,b%y)
    call rnd_switch
    atan2_st%z = atan2(a%z,b%z)
    atan2_st%acc = -1
  end function atan2_st
  !
  type (double_st) function atan2_dst(a,b) 
    implicit none
    type (double_st), intent(in) :: a, b
    if (myrandom()) call rnd_switch
    atan2_dst%x = atan2(a%x,b%x)
    if (myrandom()) call rnd_switch
    atan2_dst%y = atan2(a%y,b%y)
    call rnd_switch
    atan2_dst%z = atan2(a%z,b%z)
    atan2_dst%acc = -1
  end function atan2_dst
  !
  type (single_st) function sinh_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    sinh_st%x = sinh(a%x)
    if (myrandom()) call rnd_switch
    sinh_st%y = sinh(a%y)
    call rnd_switch
    sinh_st%z = sinh(a%z)
    sinh_st%acc = -1
  end function sinh_st
  !
  type (double_st) function sinh_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    sinh_dst%x = sinh(a%x)
    if (myrandom()) call rnd_switch
    sinh_dst%y = sinh(a%y)
    call rnd_switch
    sinh_dst%z = sinh(a%z)
    sinh_dst%acc = -1
  end function sinh_dst
!
  type (single_st) function cosh_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    cosh_st%x = cosh(a%x)
    if (myrandom()) call rnd_switch
    cosh_st%y = cosh(a%y)
    call rnd_switch
    cosh_st%z = cosh(a%z)
    cosh_st%acc = -1
  end function cosh_st
  !
  type (double_st) function cosh_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    cosh_dst%x = cosh(a%x)
    if (myrandom()) call rnd_switch
    cosh_dst%y = cosh(a%y)
    call rnd_switch
    cosh_dst%z = cosh(a%z)
    cosh_dst%acc = -1
  end function cosh_dst
!
  type (single_st) function tanh_st(a) 
    implicit none
    type (single_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_st(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    tanh_st%x = tanh(a%x)
    if (myrandom()) call rnd_switch
    tanh_st%y = tanh(a%y)
    call rnd_switch
    tanh_st%z = tanh(a%z)
    tanh_st%acc = -1
  end function tanh_st
  !
  type (double_st) function tanh_dst(a) 
    implicit none
    type (double_st), intent(in) :: a
    if (cadna_mathematic_tag) then
       if (a%acc .eq. -1) call accuracy_dst(a)
       if (a%acc.eq.0) then
          call instability(cadna_mathematic)
       end if
    end if
    if (myrandom()) call rnd_switch
    tanh_dst%x = tanh(a%x)
    if (myrandom()) call rnd_switch
    tanh_dst%y = tanh(a%y)
    call rnd_switch
    tanh_dst%z = tanh(a%z)
    tanh_dst%acc = -1
  end function tanh_dst
  !
end module cadna_math
