!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_mul
!  NAME
!    cadna_mul
!
!  DESCRIPTION
!    Defines all the functions involving at least one argument
!    of stochastic type which overload the "*" operator
!    in a statement such as "a*b".
!
!  INPUTS
!    All the functions have two arguments of stochastic, real or 
!    integer type in single or double precision. At least one 
!    argument must be of stochastic type.
!
!  RETURN VALUE
!    All the functions have one numerical output of stochastic type. 
!    It can be an array of rank 1 if one input argument is an array 
!    of rank 1. The output is in double precision if, at least,
!    one input is in double precision.
!
!  NOTES
!    If one argumment is an array of rank 1, the arguments cannot
!    be of integer type and must be in double precision.
!
!******
!
module cadna_mul
  !
  use cadna_type
  use cadna_cestac
  !
  interface operator (*)
     !       ======================
     module procedure mul_st_st, &
          &		mul_single_st, mul_st_single, mul_int_st, &
          &		mul_st_int, mul_dst_dst, mul_dst_double, &
          &		mul_double_dst, mul_dst_st, mul_st_dst, &
          &		mul_dst_single, mul_single_dst, mul_st_double, &
          &		mul_double_st, mul_int_dst, mul_dst_int, &
          !--------------------------------------------------------------
          &          mul_vdst_vdst, mul_dst_vdst, mul_vdst_dst, mul_double_vdst, &
          &          mul_vdst_double, mul_vdouble_vdst, mul_vdst_vdouble, &
          &          mul_vdouble_dst, mul_dst_vdouble, &        
          !--------------------------------------------------------------
          &          mul_dst_vvdst, mul_vvdst_dst, &
          &          mul_double_vvdst, mul_vvdst_double, &
          &          mul_double_vvvdst, mul_vvvdst_double, &
          &          mul_vvdst_single, mul_single_vvdst, &
          &          mul_vdst_single, mul_single_vdst, &
          &          mul_vvvdst_single, mul_single_vvvdst, &
          &          mul_vvdst_vvdst, &
          &          mul_int8_dst, &
          &          mul_vvvdst_vvvdst, &
          &          mul_vvvdst_dst, mul_dst_vvvdst, &
          &          mul_dst_vint
  end interface
  !
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  type (single_st) function mul_st_st(a,b)
    implicit none
    type (single_st), intent(in) :: a,b
    if (cadna_multiplication_tag) then
       if (a%acc.eq.-1) call accuracy_st(a) 
       if (a%acc.eq.0) then
          if (b%acc.eq.-1) call accuracy_st(b) 
          if (b%acc.eq.0) then
             call instability(cadna_multiplication)
          end if
       end if
    end if
    if (myrandom()) call rnd_switch
    mul_st_st%x = a%x * b%x
    if (myrandom()) call rnd_switch
    mul_st_st%y = a%y * b%y
    call rnd_switch
    mul_st_st%z = a%z * b%z
    mul_st_st%acc = -1
  end function mul_st_st
  !
  type (single_st) function mul_single_st(a,b)
    implicit none
    real, intent(in) :: a
    type (single_st), intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_single_st%x = a * b%x
    if (myrandom()) call rnd_switch
    mul_single_st%y = a * b%y
    call rnd_switch
    mul_single_st%z = a * b%z
    mul_single_st%acc = -1
  end function mul_single_st
  !
  type (single_st) function mul_st_single(a,b)
    implicit none
    type (single_st), intent(in) :: a
    real, intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_st_single%x = a%x * b
    if (myrandom()) call rnd_switch
    mul_st_single%y = a%y * b
    call rnd_switch
    mul_st_single%z = a%z * b
    mul_st_single%acc = -1
  end function mul_st_single
  !
  type (single_st) function mul_int_st(a,b)
    implicit none
    type (single_st), intent(in) :: b
    integer, intent(in) :: a
    if (myrandom()) call rnd_switch
    mul_int_st%x = a * b%x
    if (myrandom()) call rnd_switch
    mul_int_st%y = a * b%y
    call rnd_switch
    mul_int_st%z = a * b%z
    mul_int_st%acc = -1
  end function mul_int_st
  !
  type (single_st) function mul_st_int(a,b)
    implicit none
    type (single_st), intent(in) :: a
    integer, intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_st_int%x = a%x * b
    if (myrandom()) call rnd_switch
    mul_st_int%y = a%y * b
    call rnd_switch
    mul_st_int%z = a%z * b
    mul_st_int%acc = -1
  end function mul_st_int
  !
  type (double_st) function mul_dst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: a,b
    if (cadna_multiplication_tag) then
       if (a%acc.eq.-1) call accuracy_dst(a) 
       if (a%acc.eq.0) then
          if (b%acc.eq.-1) call accuracy_dst(b) 
          if (b%acc.eq.0) then
             call instability(cadna_multiplication)
          end if
       end if
    end if
    if (myrandom()) call rnd_switch
    mul_dst_dst%x = a%x * b%x
    if (myrandom()) call rnd_switch
    mul_dst_dst%y = a%y * b%y
    call rnd_switch
    mul_dst_dst%z = a%z * b%z
    mul_dst_dst%acc = -1
  end function mul_dst_dst
  !
  type (double_st) function mul_dst_double(a,b)
    implicit none
    type (double_st), intent(in) :: a
    double precision, intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_dst_double%x = a%x * b
    if (myrandom()) call rnd_switch
    mul_dst_double%y = a%y * b
    call rnd_switch
    mul_dst_double%z = a%z * b
    mul_dst_double%acc = -1
  end function mul_dst_double
  !
  type (double_st) function mul_double_dst(a,b)
    implicit none
    double precision, intent(in) :: a
    type (double_st), intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_double_dst%x = a * b%x
    if (myrandom()) call rnd_switch
    mul_double_dst%y = a * b%y
    call rnd_switch
    mul_double_dst%z = a * b%z
    mul_double_dst%acc = -1
  end function mul_double_dst
  !
  type (double_st) function mul_dst_st(a,b)
    implicit none
    type (double_st), intent(in) :: a
    type (single_st), intent(in) :: b
    if (cadna_multiplication_tag) then
       if (a%acc.eq.-1) call accuracy_dst(a) 
       if (a%acc.eq.0) then
          if (b%acc.eq.-1) call accuracy_st(b) 
          if (b%acc.eq.0) then
             call instability(cadna_multiplication)
          end if
       end if
    end if
    if (myrandom()) call rnd_switch
    mul_dst_st%x = a%x * b%x
    if (myrandom()) call rnd_switch
    mul_dst_st%y = a%y * b%y
    call rnd_switch
    mul_dst_st%z = a%z * b%z
    mul_dst_st%acc = -1
  end function mul_dst_st
  !
  type (double_st) function mul_st_dst(a,b)
    implicit none
    type (single_st), intent(in) :: a
    type (double_st), intent(in) :: b
    if (cadna_multiplication_tag) then
       if (a%acc.eq.-1) call accuracy_st(a) 
       if (a%acc.eq.0) then
          if (b%acc.eq.-1) call accuracy_dst(b) 
          if (b%acc.eq.0) then
             call instability(cadna_multiplication)
          end if
       end if
    end if
    if (myrandom()) call rnd_switch
    mul_st_dst%x = a%x * b%x
    if (myrandom()) call rnd_switch
    mul_st_dst%y = a%y * b%y
    call rnd_switch
    mul_st_dst%z = a%z * b%z
    mul_st_dst%acc = -1
  end function mul_st_dst
  !
  type (double_st) function mul_single_dst(a,b)
    implicit none
    real, intent(in) :: a
    type (double_st), intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_single_dst%x = a * b%x
    if (myrandom()) call rnd_switch
    mul_single_dst%y = a * b%y
    call rnd_switch
    mul_single_dst%z = a * b%z
    mul_single_dst%acc = -1
  end function mul_single_dst
  !
  type (double_st) function mul_dst_single(a,b)
    implicit none
    type (double_st), intent(in) :: a
    real, intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_dst_single%x = a%x * b
    if (myrandom()) call rnd_switch
    mul_dst_single%y = a%y * b
    call rnd_switch
    mul_dst_single%z = a%z * b
    mul_dst_single%acc = -1
  end function mul_dst_single
  !
  type (double_st) function mul_double_st(a,b)
    implicit none
    double precision, intent(in) :: a
    type (single_st), intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_double_st%x = a * b%x
    if (myrandom()) call rnd_switch
    mul_double_st%y = a * b%y
    call rnd_switch
    mul_double_st%z = a * b%z
    mul_double_st%acc = -1
  end function mul_double_st
  !
  type (double_st) function mul_st_double(a,b)
    implicit none
    type (single_st), intent(in) :: a
    double precision, intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_st_double%x = a%x * b
    if (myrandom()) call rnd_switch
    mul_st_double%y = a%y * b
    call rnd_switch
    mul_st_double%z = a%z * b
    mul_st_double%acc = -1
  end function mul_st_double
  !
  type (double_st) function mul_int_dst(a,b)
    implicit none
    integer, intent(in) :: a
    type (double_st), intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_int_dst%x = a * b%x
    if (myrandom()) call rnd_switch
    mul_int_dst%y = a * b%y
    call rnd_switch
    mul_int_dst%z = a * b%z
    mul_int_dst%acc = -1
  end function mul_int_dst
  !
  type (double_st) function mul_dst_int(a,b)
    implicit none
    type (double_st), intent(in) :: a
    integer, intent(in) :: b
    if (myrandom()) call rnd_switch
    mul_dst_int%x = a%x * b
    if (myrandom()) call rnd_switch
    mul_dst_int%y = a%y * b
    call rnd_switch
    mul_dst_int%z = a%z * b
    mul_dst_int%acc = -1
  end function mul_dst_int
!
!------------------------------------------------------------
!
  function mul_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a,b
    type (double_st), dimension(size(a)) :: mul_vdst_vdst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         mul_vdst_vdst(i) = mul_dst_dst(a(i), b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            mul_vdst_vdst(i)%x = a(i)%x * b(i)%x 
!         else
!            mul_vdst_vdst(i)%x = -((-a(i)%x) * b(i)%x)
!         endif
!         if (myrandom()) then
!            mul_vdst_vdst(i)%y = a(i)%y * b(i)%y
!            mul_vdst_vdst(i)%z = -((-a(i)%z) * b(i)%z)
!         else
!            mul_vdst_vdst(i)%y = -((-a(i)%y) * b(i)%y)
!            mul_vdst_vdst(i)%z = a(i)%z * b(i)%z
!        endif
!        mul_vdst_vdst(i)%acc = -1
!      enddo
!    endif
  end function mul_vdst_vdst
  !
  function mul_dst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: mul_dst_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         mul_dst_vdst(i) = mul_dst_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            mul_dst_vdst(i)%x = a%x * b(i)%x 
!         else
!            mul_dst_vdst(i)%x = -((-a%x) * b(i)%x)
!         endif
!         if (myrandom()) then
!            mul_dst_vdst(i)%y = a%y * b(i)%y
!            mul_dst_vdst(i)%z = -((-a%z) * b(i)%z)
!         else
!            mul_dst_vdst(i)%y = -((-a%y) * b(i)%y)
!            mul_dst_vdst(i)%z = a%z * b(i)%z
!         endif
!         mul_dst_vdst(i)%acc = -1
!      enddo
!    endif
  end function mul_dst_vdst
  !
  function mul_vdst_dst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: mul_vdst_dst
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         mul_vdst_dst(i) = mul_dst_dst(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            mul_vdst_dst(i)%x = a(i)%x * b%x 
!         else
!            mul_vdst_dst(i)%x = -((-a(i)%x) * b%x)
!         endif
!         if (myrandom()) then
!            mul_vdst_dst(i)%y = a(i)%y * b%y
!            mul_vdst_dst(i)%z = -((-a(i)%z) * b%z)
!         else
!            mul_vdst_dst(i)%y = -((-a(i)%y) * b%y)
!            mul_vdst_dst(i)%z = a(i)%z * b%z
!         endif
!         mul_vdst_dst(i)%acc = -1
!      enddo
!    endif
  end function mul_vdst_dst
  !
  function mul_double_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b)) :: mul_double_vdst
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         mul_double_vdst(i) = mul_double_dst(a, b(i))
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            mul_double_vdst(i)%x = a * b(i)%x 
!         else
!            mul_double_vdst(i)%x = -((-a) * b(i)%x)
!         endif
!         if (myrandom()) then
!            mul_double_vdst(i)%y = a * b(i)%y
!            mul_double_vdst(i)%z = -((-a) * b(i)%z)
!         else
!            mul_double_vdst(i)%y = -((-a) * b(i)%y)
!            mul_double_vdst(i)%z = a * b(i)%z
!         endif
!         mul_double_vdst(i)%acc = -1
!      enddo
!    endif
  end function mul_double_vdst
  !
  function mul_vdst_double(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, intent(in) :: b
    type (double_st), dimension(size(a)) :: mul_vdst_double
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         mul_vdst_double(i) = mul_dst_double(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            mul_vdst_double(i)%x = a(i)%x * b
!         else
!            mul_vdst_double(i)%x = -((-a(i)%x) * b)
!         endif
!         if (myrandom()) then
!            mul_vdst_double(i)%y = a(i)%y * b
!            mul_vdst_double(i)%z = -((-a(i)%z) * b)
!         else
!            mul_vdst_double(i)%y = -((-a(i)%y) * b)
!            mul_vdst_double(i)%z = a(i)%z * b
!         endif
!         mul_vdst_double(i)%acc = -1
!      enddo
!    endif
  end function mul_vdst_double
  !
  function mul_vdouble_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    double precision, dimension(:), intent(in) :: a
    type (double_st), dimension(size(b)) :: mul_vdouble_vdst
    integer :: k, i
    k = size(b)
!   if (cadna_cancellation_tag) then
      do i=1,k
         mul_vdouble_vdst(i) = mul_double_dst(a(i), b(i))
      enddo
!   else
!      do i=1,k
!         if (myrandom()) then
!            mul_vdouble_vdst(i)%x = a(i) * b(i)%x 
!         else
!            mul_vdouble_vdst(i)%x = -((-a(i)) * b(i)%x)
!         endif
!         if (myrandom()) then
!            mul_vdouble_vdst(i)%y = a(i) * b(i)%y
!            mul_vdouble_vdst(i)%z = -((-a(i)) * b(i)%z)
!         else
!            mul_vdouble_vdst(i)%y = -((-a(i)) * b(i)%y)
!            mul_vdouble_vdst(i)%z = a(i) * b(i)%z
!         endif
!         mul_vdouble_vdst(i)%acc = -1
!      enddo
!    endif
  end function mul_vdouble_vdst
  !
  function mul_vdst_vdouble(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: a
    double precision, dimension(:), intent(in) :: b
    type (double_st), dimension(size(a)) :: mul_vdst_vdouble
    integer :: k, i
    k = size(a)
 !   if (cadna_cancellation_tag) then
      do i=1,k
         mul_vdst_vdouble(i) = mul_dst_double(a(i), b(i))
      enddo
 !   else
 !     do i=1,k
 !        if (myrandom()) then
 !           mul_vdst_vdouble(i)%x = a(i)%x * b(i)
 !        else
 !           mul_vdst_vdouble(i)%x = -((-a(i)%x) * b(i))
 !        endif
 !        if (myrandom()) then
 !           mul_vdst_vdouble(i)%y = a(i)%y * b(i)
 !           mul_vdst_vdouble(i)%z = -((-a(i)%z) * b(i))
 !        else
 !           mul_vdst_vdouble(i)%y = -((-a(i)%y) * b(i))
!            mul_vdst_vdouble(i)%z = a(i)%z * b(i)
!         endif
!         mul_vdst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function mul_vdst_vdouble
  !
  function mul_dst_vdouble(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: mul_dst_vdouble
    integer :: k, i
    k = size(b)
!    if (cadna_cancellation_tag) then
      do i=1,k
         mul_dst_vdouble(i) = mul_dst_double(a, b(i))
      enddo
!    else  
!      do i=1,k
!         if (myrandom()) then
!            mul_dst_vdouble(i)%x = a%x * b(i)
!         else
!            mul_dst_vdouble(i)%x = -((-a%x) * b(i))
!         endif
!         if (myrandom()) then
!            mul_dst_vdouble(i)%y = a%y * b(i)
!            mul_dst_vdouble(i)%z = -((-a%z) * b(i))
!         else
!            mul_dst_vdouble(i)%y = -((-a%y) * b(i))
!            mul_dst_vdouble(i)%z = a%z * b(i)
!         endif
!         mul_dst_vdouble(i)%acc = -1
!      enddo
!    endif
  end function mul_dst_vdouble
  !
  function mul_vdouble_dst(a,b)
    implicit none
    double precision, dimension(:), intent(in) :: a
    type (double_st), intent(in) :: b
    type (double_st), dimension(size(a)) :: mul_vdouble_dst
    integer :: k, i
    k = size(a)
!    if (cadna_cancellation_tag) then
      do i=1,k
         mul_vdouble_dst(i) = mul_double_dst(a(i), b)
      enddo
!    else
!      do i=1,k
!         if (myrandom()) then
!            mul_vdouble_dst(i)%x = a(i) * b%x 
!         else
!            mul_vdouble_dst(i)%x = -((-a(i)) * b%x)
!         endif
!         if (myrandom()) then
!            mul_vdouble_dst(i)%y = a(i) * b%y
!            mul_vdouble_dst(i)%z = -((-a(i)) * b%z)
!         else
!            mul_vdouble_dst(i)%y = -((-a(i)) * b%y)
!            mul_vdouble_dst(i)%z = a(i) * b%z
!         endif
!         mul_vdouble_dst(i)%acc = -1
!      enddo
!    endif
  end function mul_vdouble_dst
  !
  function mul_dst_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2)) :: mul_dst_vvdst
    integer i, n
    n=size(b,2)
    do i=1,n
       mul_dst_vvdst(:,i)=mul_dst_vdst(a,mul_dst_vvdst(:,i))
    enddo
  end function mul_dst_vvdst
  !
  function mul_vvdst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: b
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2)) :: mul_vvdst_dst
    mul_vvdst_dst=mul_dst_vvdst(b,a)
  end function mul_vvdst_dst
  !
  function mul_double_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2)) :: mul_double_vvdst
    integer i, n
    n=size(b,2)
    do i=1,n
       mul_double_vvdst(:,i)=mul_double_vdst(a,mul_double_vvdst(:,i))
    enddo
  end function mul_double_vvdst
  !
  function mul_vvdst_double(a,b)
    implicit none
    double precision, intent(in) :: b
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2)) :: mul_vvdst_double
    mul_vvdst_double=mul_double_vvdst(b,a)
  end function mul_vvdst_double
  !
  function mul_double_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    double precision, intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2), size(b,3)) :: mul_double_vvvdst
    integer i, j, k
    do k=1,size(b,3)
       do j=1,size(b,2)
          do i=1,size(b,1)
             mul_double_vvvdst(i,j,k)=mul_double_dst(a,b(i,j,k))
          enddo
       enddo
    enddo
  end function mul_double_vvvdst
  !
  function mul_vvvdst_double(a,b)
    implicit none
    double precision, intent(in) :: b
    type (double_st), dimension(:,:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2), size(a,3)) :: mul_vvvdst_double
    mul_vvvdst_double=mul_double_vvvdst(b,a)
  end function mul_vvvdst_double
  !
  function mul_vvdst_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2)) :: mul_vvdst_vvdst
    integer i,j
    do j=1,size(a,2)
       do i=1,size(a,1)
          mul_vvdst_vvdst(i,j)=mul_dst_dst(a(i,j),b(i,j))
       enddo
    enddo
  end function mul_vvdst_vvdst
  !
  function mul_vvdst_single(a,b)
    implicit none
    real, intent(in) :: b
    type (double_st), dimension(:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2)) :: mul_vvdst_single
    double precision :: c
    c=b
    mul_vvdst_single=mul_vvdst_double(a,c)
  end function mul_vvdst_single
  !
  function mul_single_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(in) :: b
    real, intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2)) :: mul_single_vvdst
    mul_single_vvdst=mul_vvdst_single(b,a)
  end function mul_single_vvdst
  !
  function mul_vdst_single(a,b)
    implicit none
    real, intent(in) :: b
    type (double_st), dimension(:), intent(in) :: a
    type (double_st), dimension(size(a)) :: mul_vdst_single
    double precision :: c
    c=b
    mul_vdst_single=mul_vdst_double(a,c)
  end function mul_vdst_single
  !
  function mul_single_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(in) :: b
    real, intent(in) :: a
    type (double_st), dimension(size(b)) :: mul_single_vdst
    mul_single_vdst=mul_vdst_single(b,a)
  end function mul_single_vdst
  !
  function mul_vvvdst_single(a,b)
    implicit none
    real, intent(in) :: b
    type (double_st), dimension(:,:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2), size(a,3)) :: mul_vvvdst_single
    double precision :: c
    c=b
    mul_vvvdst_single=mul_vvvdst_double(a,c)
  end function mul_vvvdst_single
  !
  function mul_single_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    real, intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2), size(b,3)) :: mul_single_vvvdst
    mul_single_vvvdst=mul_vvvdst_single(b,a)
  end function mul_single_vvvdst
  !
  function mul_int8_dst(a,b)
    implicit none
    type (double_st), intent(in) :: b
    integer*8, intent(in) :: a
    type (double_st) :: mul_int8_dst
    double precision :: c
    c=a
    mul_int8_dst=mul_double_dst(c,b)
  end function mul_int8_dst
  !
  function mul_vvvdst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    type (double_st), dimension(:,:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2), size(a,3)) :: mul_vvvdst_vvvdst
    integer i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             mul_vvvdst_vvvdst(i,j,k)=mul_dst_dst(a(i,j,k),b(i,j,k))
          enddo
       enddo
    enddo
  end function mul_vvvdst_vvvdst
  !
  function mul_vvvdst_dst(a,b)
    implicit none
    type (double_st), intent(in) :: b
    type (double_st), dimension(:,:,:), intent(in) :: a
    type (double_st), dimension(size(a,1), size(a,2), size(a,3)) :: mul_vvvdst_dst
    integer i,j,k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             mul_vvvdst_dst(i,j,k)=mul_dst_dst(a(i,j,k),b)
          enddo
       enddo
    enddo
    
  end function mul_vvvdst_dst
  !
  function mul_dst_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b,1), size(b,2), size(b,3)) :: mul_dst_vvvdst
    mul_dst_vvvdst=mul_vvvdst_dst(b,a)
  end function mul_dst_vvvdst
  !
  function mul_dst_vint(a,b)
    implicit none
    integer, dimension(:), intent(in) :: b
    type (double_st), intent(in) :: a
    type (double_st), dimension(size(b)) :: mul_dst_vint
    integer i
    do i=1,size(b)
       mul_dst_vint(i)=mul_dst_int(a,b(i))
    enddo
  end function mul_dst_vint
  !
end module cadna_mul
