!Copyright 2016 J.-M. CHESNEAUX, F. JEZEQUEL, J.-L. LAMOTTE
!
!This file is part of CADNA.
!
!    CADNA is free software: you can redistribute it and/or modify it
!    under the terms of the GNU Lesser General Public License as
!    published by the Free Software Foundation, either version 3 of the
!    License, or (at your option) any later version.
!
!    CADNA is distributed in the hope that it will be useful, but WITHOUT
!    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
!    Public License for more details.
!
!    You should have received a copy of the GNU Lesser General Public
!    License along with CADNA. If not, see
!    <http://www.gnu.org/licenses/>.
!
!****h module cadna_to
!  NAME
!    cadna_to
!
!  DESCRIPTION
!    Defines all the functions involving at least one argument
!    of stochastic type which overload the assignment
!    statement "=".
!
!  INPUTS
!    All the functions have one "inout" argument and one "in" 
!    argument. At least one argument must be of stochastic type.
!    Arguments can be arrays of rank 1.
!
!  OUTPUTS
!    The same variable that is "inout" in the "input" argument.
!
!  NOTES
!    If one argumment is an array of rank 1, the arguments cannot
!    be of integer type and must be in double precision.
!
!******
!
module cadna_to
  !
  use cadna_type
  !
  !
  interface assignment (=)
     !	======================
     module procedure &
          &          to_st_single, to_single_st, to_int_st, to_st_int, &
          &		to_dst_double, to_double_dst, &
          &		to_dst_st, to_st_dst, to_dst_single, to_single_dst, &
          &		to_st_double, to_double_st, to_int_dst, to_dst_int, &
          !--------------------------------------------------------------
          &          to_vdst_vdst, to_dst_vdst, to_double_vdst, to_vdouble_vdst, to_dst_vdouble, &
          &          to_vdst_vdouble, &
          !--------------------------------------------------------------
          &          to_single_vdst, &
          &          to_int_vdst, &
          &          to_double_vvdst, &
          &          to_single_vvdst, &
          &          to_vdst_vsingle, &
          &          to_vvdouble_vvdst, &
          &          to_vvvdouble_vvvdst, &
          &          to_int_vvdst, &
          &          to_int_vvvdst, &
          &          to_double_vvvdst, &
          &          to_int_vvvvdst, &
          &          to_vdst_vint, &
          &          to_vint_vdst, &
          &          to_vsingle_vdst, &
          &          to_int8_dst, to_dst_int8, &
          &          to_dst_vvint, &
          &          to_double_vvvvdst
  end interface
  
  !
!!!!!!!!!!!!!!!!!!!!!
contains   !!
!!!!!!!!!!!!!!!!!!!!!
  !
  !
  subroutine to_single_st(a,b)
    implicit none
    type (single_st), intent(out) :: a
    real, intent(in) :: b
    a%x = b
    a%y = b
    a%z = b
    a%acc = 7
  end subroutine to_single_st
  !
  subroutine to_st_single(a,b)
    implicit none
    real, intent(out) :: a
    type (single_st), intent(in) :: b
    a = (b%x + b%y + b%z) / 3.0
  end subroutine to_st_single
  !
  subroutine to_int_st(a,b)
    implicit none
    type (single_st), intent(out) :: a
    integer, intent(in) :: b
    if (myrandom()) call rnd_switch
    a%x = b
    if (myrandom()) call rnd_switch
    a%y = b
    call rnd_switch
    a%z = b
    a%acc = 7
  end subroutine to_int_st
  !
  subroutine to_st_int(a,b)
    implicit none
    integer, intent(out) :: a
    type (single_st), intent(in) :: b
    a = (b%x + b%y + b%z) / 3.0
  end subroutine to_st_int
  !
  subroutine to_dst_double(a,b)
    implicit none
    double precision, intent(out) :: a
    type (double_st), intent(in) :: b
    a = (b%x + b%y + b%z) / 3.0
  end subroutine to_dst_double
  !
  subroutine to_double_dst(a,b)
    implicit none
    type (double_st), intent(out) :: a
    double precision, intent(in) :: b
    a%x = b
    a%y = b
    a%z = b
    a%acc = 15
  end subroutine to_double_dst
  !
  subroutine to_dst_st(a,b)
    implicit none
    type (single_st), intent(out) :: a
    type (double_st), intent(in) :: b
    if (myrandom()) call rnd_switch
    a%x = b%x
    if (myrandom()) call rnd_switch
    a%y = b%y
    call rnd_switch
    a%z = b%z
    a%acc = min(7,b%acc)
  end subroutine to_dst_st
  !
  subroutine to_st_dst(a,b)
    implicit none
    type (double_st), intent(out) :: a
    type (single_st), intent(in) :: b
    a%x = b%x
    a%y = b%y
    a%z = b%z
    a%acc = b%acc
  end subroutine to_st_dst
  !
  !if a dst variable is assigned from a real, 
  !it is pertubred with the relative error 1.E-7.
  subroutine to_single_dst(a,b)
    implicit none
    type (double_st), intent(out) :: a
    real, intent(in) :: b
    real :: aux
    a%x = b
    call rnd_plinf
    call random_number(aux)
    a%x = a%x*(1.0 + aux*1.D-7)
    a%y = b
    call rnd_moinf
    call random_number(aux)
    a%y = a%y*(1.0 - aux*1.D-7)
    a%z = b
    a%acc = -1
  end subroutine to_single_dst
  !
  subroutine to_dst_single(a,b)
    implicit none
    real, intent(out) :: a
    type (double_st), intent(in) :: b
    a = (b%x + b%y + b%z) / 3.0
  end subroutine to_dst_single
  !
  subroutine to_st_double(a,b)
    implicit none
    double precision, intent(out) :: a
    type (single_st), intent(in) :: b
    a = (b%x + b%y + b%z) / 3.0
  end subroutine to_st_double
  !
  subroutine to_double_st(a,b)
    implicit none
    type (single_st), intent(out) :: a
    double precision, intent(in) :: b
    if (myrandom()) call rnd_switch
    a%x = b
    if (myrandom()) call rnd_switch
    a%y = b
    call rnd_switch
    a%z = b
    a%acc = 7
  end subroutine to_double_st
  !
  subroutine to_int_dst(a,b)
    implicit none
    type (double_st), intent(out) :: a
    integer, intent(in) :: b
    if (myrandom()) call rnd_switch
    a%x = b
    if (myrandom()) call rnd_switch
    a%y = b
    call rnd_switch
    a%z = b
    a%acc = 15
  end subroutine to_int_dst
  !
  subroutine to_dst_int(a,b)
    implicit none
    integer, intent(out) :: a
    type (double_st), intent(in) :: b
    a = (b%x + b%y + b%z) / 3.0
  end subroutine to_dst_int
  !
  !------------------------------------------------------------
  !
  subroutine to_vdst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    type (double_st), dimension(:), intent(in) :: b
    a%x = b%x
    a%y = b%y
    a%z = b%z
    a%acc = b%acc
  end subroutine to_vdst_vdst
  !
  subroutine to_dst_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    type (double_st), intent(in) :: b
    a%x = b%x
    a%y = b%y
    a%z = b%z
    a%acc = b%acc
  end subroutine to_dst_vdst
  !
  subroutine to_double_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    double precision, intent(in) :: b
    a%x = b
    a%y = b
    a%z = b
    a%acc = 15
  end subroutine to_double_vdst
  !
  subroutine to_vdouble_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    double precision, dimension(:), intent(in) :: b
    a%x = b
    a%y = b
    a%z = b
    a%acc = 15
  end subroutine to_vdouble_vdst
  !
  subroutine to_dst_vdouble(a,b)
    implicit none
    double precision, dimension(:), intent(out) :: a
    type (double_st), intent(in) :: b
    a = (b%x + b%y + b%z)/3.0
  end subroutine to_dst_vdouble
  !
  subroutine to_vdst_vdouble(a,b)
    implicit none
    double precision, dimension(:), intent(out) :: a
    type (double_st), dimension(:), intent(in) :: b
    a = (b%x + b%y + b%z)/3.0
  end subroutine to_vdst_vdouble
  !
  subroutine to_single_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    real, intent(in) :: b
    integer :: i
    do i=1,size(a)
       call to_single_dst(a(i),b)
    enddo
  end subroutine to_single_vdst
  !
  subroutine to_int_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    integer, intent(in) :: b
    integer :: i
    do i=1,size(a)
       call to_int_dst(a(i),b)
    enddo
  end subroutine to_int_vdst
  !
  subroutine to_double_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(out) :: a
    double precision, intent(in) :: b
    integer :: i, j
    do j=1,size(a,2)
       do i=1,size(a,1)
          call to_double_dst(a(i,j),b)
       enddo
    enddo
  end subroutine to_double_vvdst
  !
  subroutine to_single_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(out) :: a
    real, intent(in) :: b
    integer :: i, j
    do j=1,size(a,2)
       do i=1,size(a,1)
          call to_single_dst(a(i,j),b)
       enddo
    enddo
  end subroutine to_single_vvdst
  !
  subroutine to_vdst_vsingle(a,b)
    implicit none
    real, dimension(:), intent(out) :: a
    type (double_st), dimension(:), intent(in) :: b
    a = (b%x + b%y + b%z)/3.0
  end subroutine to_vdst_vsingle
  !
  subroutine to_vvdouble_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(out) :: a
    double precision, dimension(:,:), intent(in) :: b
    integer :: i, j
    do j=1,size(a,2)
       do i=1,size(a,1)
          call to_double_dst(a(i,j),b(i,j))
       enddo
    enddo
  end subroutine to_vvdouble_vvdst
  !
  subroutine to_vvvdouble_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(out) :: a
    double precision, dimension(:,:,:), intent(in) :: b
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             call to_double_dst(a(i,j,k),b(i,j,k))
          enddo
       enddo
    enddo
  end subroutine to_vvvdouble_vvvdst
  !
  subroutine to_int_vvdst(a,b)
    implicit none
    type (double_st), dimension(:,:), intent(out) :: a
    integer, intent(in) :: b
    integer :: i, j
    do j=1,size(a,2)
       do i=1,size(a,1)
          call to_int_dst(a(i,j),b)
       enddo
    enddo
  end subroutine to_int_vvdst
  !
  subroutine to_int_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(out) :: a
    integer, intent(in) :: b
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             call to_int_dst(a(i,j,k),b)
          enddo
       enddo
    enddo
  end subroutine to_int_vvvdst
  !
  subroutine to_double_vvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:), intent(out) :: a
    double precision, intent(in) :: b
    integer :: i, j, k
    do k=1,size(a,3)
       do j=1,size(a,2)
          do i=1,size(a,1)
             call to_double_dst(a(i,j,k),b)
          enddo
       enddo
    enddo
  end subroutine to_double_vvvdst
  !
  subroutine to_int_vvvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:,:), intent(out) :: a
    integer, intent(in) :: b
    integer :: i, j, k, l
    do l=1,size(a,4)
       do k=1,size(a,3)
          do j=1,size(a,2)
             do i=1,size(a,1)
                call to_int_dst(a(i,j,k,l),b)
             enddo
          enddo
       enddo
    enddo
  end subroutine to_int_vvvvdst
  !
  subroutine to_vdst_vint(a,b)
    implicit none
    integer, dimension(:), intent(out) :: a
    type (double_st), dimension(:), intent(in) :: b
    double precision :: c
    integer :: i
    do i=1,size(a)
       c=b(i)
       a(i)=c
    enddo
  end subroutine to_vdst_vint
  !
  subroutine to_vint_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    integer, dimension(:), intent(in) :: b
    double precision :: c
    integer :: i
    do i=1,size(a)
       a(i)=b(i)
    enddo
  end subroutine to_vint_vdst
  !
  subroutine to_vsingle_vdst(a,b)
    implicit none
    type (double_st), dimension(:), intent(out) :: a
    real, dimension(:), intent(in) :: b
    integer :: i
    do i=1,size(a)
       a(i)=b(i)
    enddo
  end subroutine to_vsingle_vdst
  !
  subroutine to_int8_dst(a,b)
    implicit none
    type (double_st), intent(out) :: a
    integer*8, intent(in) :: b
    double precision :: c
    c=b
    call to_double_dst(a,c)
  end subroutine to_int8_dst
  !
  subroutine to_dst_int8(a,b)
    implicit none
    integer*8, intent(out) :: a
    type (double_st), intent(in) :: b
    double precision :: c
    call to_dst_double(c,b)
    a=c
  end subroutine to_dst_int8
  !
  subroutine to_dst_vvint(a,b)
    implicit none
    integer, dimension(:,:), intent(out) :: a
    type (double_st), intent(in) :: b
    double precision :: c
    integer :: i,j
    call to_dst_double(c,b)
    do j=1,size(a,2)
       do i=1,size(a,1)
          a(i,j)=c
       enddo
    enddo
  end subroutine to_dst_vvint
  !
  subroutine to_double_vvvvdst(a,b)
    implicit none
    type (double_st), dimension(:,:,:,:), intent(out) :: a
    double precision, intent(in) :: b
    integer :: i, j, k, l
    do l=1,size(a,4)
       do k=1,size(a,3)
          do j=1,size(a,2)
             do i=1,size(a,1)
                call to_double_dst(a(i,j,k,l),b)
             enddo
          enddo
       enddo
    enddo
  end subroutine to_double_vvvvdst
  !
end module cadna_to
